# -*- coding: utf-8 -*

import getopt
import os
import sys

from east import applications
from east import formatting
from east.synonyms import synonyms
from east import utils


def main():
    args = sys.argv[1:]
    opts, args = getopt.getopt(args, "a:f:l:t:ds")
    opts = dict(opts)
    opts.setdefault("-a", "easa")   # Algorithm to use for computing ASTs
    opts.setdefault("-l", "0.6")    # Level of significance for graph construction
    opts.setdefault("-t", "0.25")   # Threshold of the matching score
    # NOTE(msdubov): -f (output format) option takes different values for different
    #                subcommands and its default value is set in corresponding handlers.

    if len(args) < 2:
        print("Invalid syntax: EAST should be called as:\n\n"
              "    east <command> <subcommand> [options] args\n\n"
              "Commands available: keyphrases.\n"
              "Subcommands available: table/graph.")
        return 1

    command = args[0]
    subcommand = args[1]

    if command == "keyphrases":

        if len(args) < 4:
            print('Invalid syntax. For keyphrases analysis, EAST should be called as:\n\n'
                  '    east keyphrases <subcommand> [options] "path/to/keyphrases.txt" '
                  '"path/to/texts/dir"')
            return 1

        keyphrases_file = os.path.abspath(args[2])
        input_path = os.path.abspath(args[3])
        use_synonyms = "-s" in opts
        normalized_scores = "-d" not in opts
        ast_algorithm = opts["-a"]
        significance_level = float(opts["-l"])
        score_threshold = float(opts["-t"])

        if os.path.isdir(input_path):
            input_files = [os.path.abspath(input_path) + "/" + filename
                           for filename in os.listdir(input_path)
                           if filename.endswith(".txt")]
        else:
            input_files = [os.path.abspath(input_path)]

        texts = {}
        for filename in input_files:
            with open(filename) as f:
                text_name = os.path.basename(filename).decode("utf-8")[:-4]
                texts[text_name] = f.read()

        with open(keyphrases_file) as f:
            keyphrases = map(lambda k: utils.prepare_text(k), f.read().splitlines())

        if use_synonyms:
            synonimizer = synonyms.SynonymExtractor(input_path)
        else:
            synonimizer = None

        if subcommand == "table":

            keyphrases_table = applications.keyphrases_table(keyphrases, texts, ast_algorithm,
                                                             normalized_scores, synonimizer)
            opts.setdefault("-f", "xml")  # Table output format (also "csv" possible)
            table_format = opts["-f"].lower()

            if table_format == "xml":
                res = formatting.table2xml(keyphrases_table)
            elif table_format == "csv":
                res = formatting.table2csv(keyphrases_table)
            else:
                print ("Unknown table format: '%s'. "
                       "Please use one of: 'xml', 'csv'." % table_format)
                return 1

            print res.encode("utf-8", "ignore")

        elif subcommand == "graph":

            graph = applications.keyphrases_graph(keyphrases, texts, significance_level,
                                                  score_threshold, ast_algorithm,
                                                  normalized_scores, synonimizer)

            opts.setdefault("-f", "edges")  # Graph output format (also "gml" possible)
            graph_format = opts["-f"].lower()

            if graph_format == "gml":
                res = formatting.graph2gml(graph)
            elif graph_format == "edges":
                res = formatting.graph2edges(graph)
            else:
                print ("Unknown graph format: '%s'. "
                       "Please use one of: 'gml', 'edges'." % graph_format)
                return 1

            print res.encode("utf-8", "ignore")

        else:
            print "Invalid subcommand: '%s'. Please use one of: 'table', 'graph'." % subcommand
            return 1

    else:
        print "Invalid command: '%s'. Please use one of: 'keyphrases'." % command
        return 1



if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*

from east.asts import ast
from east.asts import utils
from east import consts

class LinearAnnotatedSuffixTree(ast.AnnotatedSuffixTree):

    __algorithm__ = "ast_linear"

    def _construct(self, strings_collection):
        """
        Generalized suffix tree construction algorithm based on the
        Ukkonen's algorithm for suffix tree construction,
        with linear [O(n_1 + ... + n_m)] worst-case time complexity,
        where m is the number of strings in collection.
        
        """
        
        # 1. Add a unique character to each string in the collection
        strings_collection = utils.make_unique_endings(strings_collection)
    
        ############################################################
        # 2. Build the GST using modified Ukkonnen's algorithm     #
        ############################################################
        
        root = ast.AnnotatedSuffixTree.Node()
        root.strings_collection = strings_collection
        # To preserve simplicity
        root.suffix_link = root
        root._arc = (0,-1,0)
        # For constant updating of all leafs, see [Gusfield {RUS}, p. 139]
        root._e = [0 for _ in xrange(len(strings_collection))]
        
        
        def _ukkonen_first_phases(string_ind):
            """
            Looks for the part of the string which is already encoded.
            Returns a tuple of form
            ([length of already encoded string preffix],
             [tree node to start the first explicit phase with],
             [path to go down at the beginning of the first explicit phase]).
            
            """
            already_in_tree = 0
            suffix = strings_collection[string_ind]
            starting_path = (0, 0, 0)
            starting_node = root
            child_node = starting_node.chose_arc(suffix)
            while child_node:
                (str_ind, substr_start, substr_end) = child_node.arc()
                match = utils.match_strings(
                            suffix, strings_collection[str_ind][substr_start:substr_end])
                already_in_tree += match
                if match == substr_end-substr_start:
                    # matched the arc, proceed with child node
                    suffix = suffix[match:]
                    starting_node = child_node
                    child_node = starting_node.chose_arc(suffix)
                else:
                    # otherwise we will have to proceed certain path at the beginning
                    # of the first explicit phase
                    starting_path = (str_ind, substr_start, substr_start+match)
                    break
            # For constant updating of all leafs, see [Gusfield {RUS}, p. 139]
            root._e[string_ind] = already_in_tree
                
            return (already_in_tree, starting_node, starting_path)
        
        def _ukkonen_phase(string_ind, phase, starting_node, starting_path, starting_continuation):
            """
            Ukkonen's algorithm single phase.
            Returns a tuple of form:
            ([tree node to start the next phase with],
             [path to go down at the beginning of the next phase],
             [starting continuation for the next phase]).
            
            """
            current_suffix_end = starting_node
            suffix_link_source_node = None
            path_str_ind, path_substr_start, path_substr_end = starting_path
            # Continuations [starting_continuation..(i+1)]
            for continuation in xrange(starting_continuation, phase+1):
                # Go up to the first node with suffix link [no more than 1 pass]
                if continuation > starting_continuation:
                    path_str_ind, path_substr_start, path_substr_end = 0, 0, 0
                    if not current_suffix_end.suffix_link:
                        (path_str_ind, path_substr_start, path_substr_end) = current_suffix_end.arc()
                        current_suffix_end = current_suffix_end.parent
                    if current_suffix_end.is_root():
                        path_str_ind = string_ind
                        path_substr_start = continuation
                        path_substr_end = phase
                    else:
                        # Go through the suffix link
                        current_suffix_end = current_suffix_end.suffix_link
                        
                # Go down the path (str_ind, substr_start, substr_end)
                # NB: using Skip/Count trick,
                # see [Gusfield {RUS} p.134] for details
                g = path_substr_end - path_substr_start
                if g > 0:
                    current_suffix_end = current_suffix_end.chose_arc(strings_collection
                                         [path_str_ind][path_substr_start])
                (_, cs_ss_start, cs_ss_end) = current_suffix_end.arc()
                g_ = cs_ss_end - cs_ss_start
                while g >= g_:
                    path_substr_start += g_
                    g -= g_
                    if g > 0:
                        current_suffix_end = current_suffix_end.chose_arc(strings_collection
                                             [path_str_ind][path_substr_start])
                    (_, cs_ss_start, cs_ss_end) = current_suffix_end.arc()
                    g_ = cs_ss_end - cs_ss_start
                    
                # Perform continuation by one of three rules,
                # see [Gusfield {RUS} p. 129] for details
                if g == 0:
                    # Rule 1
                    if current_suffix_end.is_leaf():
                        pass
                    # Rule 2a
                    elif not current_suffix_end.chose_arc(strings_collection[string_ind][phase]):
                        if suffix_link_source_node:
                            suffix_link_source_node.suffix_link = current_suffix_end
                        new_leaf = current_suffix_end.add_new_child(string_ind, phase, -1)
                        new_leaf.weight = 1
                        if continuation == starting_continuation:
                            starting_node = new_leaf
                            starting_path = (0, 0, 0)
                    # Rule 3a
                    else:
                        if suffix_link_source_node:
                            suffix_link_source_node.suffix_link = current_suffix_end
                        starting_continuation = continuation
                        starting_node = current_suffix_end
                        starting_path = (string_ind, phase, phase+1)
                        break
                    suffix_link_source_node = None
                else:
                    (si, ss, se) = current_suffix_end._arc
                    # Rule 2b
                    if strings_collection[si][ss + g] != strings_collection[string_ind][phase]:
                        parent = current_suffix_end.parent
                        parent.remove_child(current_suffix_end)
                        current_suffix_end._arc = (si, ss+g, se)
                        new_node = parent.add_new_child(si, ss, ss + g)
                        new_leaf = new_node.add_new_child(string_ind, phase, -1)
                        new_leaf.weight = 1
                        if continuation == starting_continuation:
                            starting_node = new_leaf
                            starting_path = (0, 0, 0)
                        new_node.add_child(current_suffix_end)
                        if suffix_link_source_node:
                            # Define new suffix link
                            suffix_link_source_node.suffix_link = new_node
                        suffix_link_source_node = new_node
                        current_suffix_end = new_node
                    # Rule 3b
                    else:
                        suffix_link_source_node = None
                        starting_continuation = continuation
                        starting_node = current_suffix_end.parent
                        starting_path = (si, ss, ss+g+1)
                        break
            
            # Constant updating of all leafs, see [Gusfield {RUS}, p. 139]
            starting_node._e[string_ind] += 1
            
            return starting_node, starting_path, starting_continuation
                        
        for m in xrange(len(strings_collection)):
            # Check for phases 1..x that are already in tree
            starting_phase, starting_node, starting_path = _ukkonen_first_phases(m)
            starting_continuation = 0
            # Perform phases (x+1)..n explicitly
            for phase in xrange(starting_phase, len(strings_collection[m])):
                starting_node, starting_path, starting_continuation = \
                    _ukkonen_phase(m, phase, starting_node, starting_path, starting_continuation)
                    
        
        ############################################################
        ############################################################
        ############################################################
        
        # 3. Delete degenerate first-level children
        for k in root.children.keys():
            (ss, si, se) = root.children[k].arc()
            if (se - si == 1 and
                ord(strings_collection[ss][si]) >= consts.String.UNICODE_SPECIAL_SYMBOLS_START):
                del root.children[k]
        
        # 4. Make a depth-first bottom-up traversal and annotate
        #    each node by the sum of its children;
        #    each leaf is already annotated with '1'.
        def _annotate(node):
            weight = 0
            for k in node.children:
                if node.children[k].weight > 0:
                    weight += node.children[k].weight
                else:
                    weight += _annotate(node.children[k])
            node.weight = weight
            return weight
        _annotate(root)
        
        return root



# -*- coding: utf-8 -*

import testtools

from east import utils


class UtilsTestCase(testtools.TestCase):

    def test_tokenize(self):
        text = "Well, what a sunny day!"
        tokens = ["Well", "what", "a", "sunny", "day"]
        self.assertEqual(utils.tokenize(text), tokens)



# -*- coding: utf-8 -*

import os
import random
import re
import sys


class ImmutableMixin(object):
    _inited = False

    def __init__(self):
        self._inited = True

    def __setattr__(self, key, value):
        if self._inited:
            raise exceptions.ImmutableException()
        super(ImmutableMixin, self).__setattr__(key, value)


class EnumMixin(object):
    def __iter__(self):
        for k, v in itertools.imap(lambda x: (x, getattr(self, x)), dir(self)):
            if not k.startswith('_'):
                yield v


def prepare_text(text):
    text = unicode(text.decode('utf-8', errors='replace'))
    text = text.upper()
    return text


def tokenize(text):
    return re.findall(re.compile("[\w']+", re.U), text)


def text_to_strings_collection(text, words=3):
    '''
    Splits the text to a collection of strings;
    a GAST for such a split collection usually produces
    better results in keyword matching that a GAST
    for the whole text. The word parameters determines
    how many words each string in the collection shall
    consist of (3 by default)
    
    return: Unicode
    '''
    
    text = prepare_text(text)
    strings_collection = tokenize(text)
    strings_collection = filter(lambda s: len(s) > 2 and not s.isdigit(), strings_collection)
        
    i = 0
    strings_collection_grouped = []
    while i < len(strings_collection):
        group = ''
        for j in xrange(words):
            if i + j < len(strings_collection):
                group += strings_collection[i+j]
        strings_collection_grouped.append(group)
        i += words

    # Having an empty strings collection would lead to a runtime errors in the applications.
    if not strings_collection_grouped:
        strings_collection_grouped = [" "]
        
    return strings_collection_grouped


def random_string(length):
    string = u"".join([unichr(ord("A") + random.randint(0, 25)) for _ in xrange(length - 2)])
    return string


def output_is_redirected():
    return os.fstat(0) != os.fstat(1)


def itersubclasses(cls, _seen=None):
    """Generator over all subclasses of a given class in depth first order."""

    if not isinstance(cls, type):
        raise TypeError(_('itersubclasses must be called with '
                          'new-style classes, not %.100r') % cls)
    _seen = _seen or set()
    try:
        subs = cls.__subclasses__()
    except TypeError:   # fails only when cls is type
        subs = cls.__subclasses__(cls)
    for sub in subs:
        if sub not in _seen:
            _seen.add(sub)
            yield sub
            for sub in itersubclasses(sub, _seen):
                yield sub


def import_modules_from_package(package):
    """Import modules from package and append into sys.modules

    :param package: full package name, e.g. east.asts
    """
    path = [os.path.dirname(__file__), '..'] + package.split('.')
    path = os.path.join(*path)
    for root, dirs, files in os.walk(path):
        for filename in files:
            if filename.startswith('__') or not filename.endswith('.py'):
                continue
            new_package = ".".join(root.split(os.sep)).split("....")[1]
            module_name = '%s.%s' % (new_package, filename[:-3])
            if module_name not in sys.modules:
                __import__(module_name)



# -*- coding: utf-8 -*

import getopt
import sys

from east.synonyms import synonyms

def main(args):
    opts, args = getopt.getopt(args, "")

    path = args[0]

    synonimizer = synonyms.SynonymExtractor(path)
    print "Prepared synonimizer\n"

    synonym_dicts = synonimizer.get_synonyms(threshold=0.3, return_similarity_measure=True)
    for w1 in synonym_dicts:
        syn_str = ["%s (%.2f)" % (w2, sim) for w2, sim in synonym_dicts[w1]]
        print "%s -> %s" % (w1, ", ".join(syn_str))

if __name__ == "__main__":
    main(sys.argv[1:])



# -*- coding: utf-8 -*

import codecs
import collections
import itertools
import math
import os
import subprocess
from xml.dom import minidom


from east import consts
from east import exceptions
from east import utils as common_utils
from east.synonyms import utils


class SynonymExtractor(object):

    def __init__(self, input_path):
        self.current_os = utils.determine_operating_system()
        self.tomita_path, self.tomita_binary = self._get_tomita_path()
        if self.tomita_binary is None:
            raise exceptions.TomitaNotInstalledException()
        self.text, self.number_of_texts = self._retrieve_text(input_path)
        self.dependency_triples, self.dt_for_r, self.dt_for_w1r, self.dt_for_rw2 = \
            self._retrieve_dependency_triples(self.text)
        self.word_frequencies = self._calculate_word_frequencies(self.text)
        self.frequencies = self._calculate_dt_frequencies(self.dependency_triples)
        self.words = set([dt[0] for dt in self.dependency_triples] +
                         [dt[2] for dt in self.dependency_triples])
        self.relations = set([dt[1] for dt in self.dependency_triples])
        self.I_memoized = {}
        self.T_memoized = {}
        self.synonyms_memoized = {}

    def _retrieve_text(self, input_path):
        if os.path.isdir(input_path):
            text = ""
            number_of_texts = 0
            for file_name in os.listdir(input_path):
                if file_name.endswith(".txt"):
                    with open(os.path.abspath(input_path) + "/" + file_name) as f:
                        text += f.read()
                        number_of_texts += 1
        else:
            with open(input_path) as f:
                text = f.read()
                number_of_texts = 1
        return text, number_of_texts

    def _retrieve_dependency_triples(self, text):

        dependency_triples = []

        # Additional indexes to speed up the calculation of I(w1, r, w2)
        dt_for_r = collections.defaultdict(list)
        dt_for_w1r = collections.defaultdict(list)
        dt_for_rw2 = collections.defaultdict(list)

        p = subprocess.Popen([self.tomita_binary, "config.proto"],
                             stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE, cwd=self.tomita_path,
                             shell=(self.current_os == consts.OperatingSystem.WINDOWS))
        out, err = p.communicate(input=text)

        xmldoc = minidom.parseString(out)
        relations = xmldoc.getElementsByTagName('Relation') 
        
        for rel in relations:
            r = rel.childNodes[0].nodeName
            w1, w2 = rel.childNodes[0].attributes['val'].value.split(' ', 1)

            dt = (w1, r, w2)
            dependency_triples.append(dt)
            dt_for_r[r].append(dt)
            dt_for_w1r[(w1, r)].append(dt)
            dt_for_rw2[(r, w2)].append(dt)

            # NOTE(msdubov): Also add inversed triples.
            r_inv = r[:-3] if r.endswith("_of") else (r + "_of")
            dt_inv = (w2, r_inv, w1)
            dependency_triples.append(dt_inv)
            dt_for_r[r_inv].append(dt_inv)
            dt_for_w1r[(w2, r_inv)].append(dt_inv)
            dt_for_rw2[(r_inv, w1)].append(dt_inv)

        return dependency_triples, dt_for_r, dt_for_w1r, dt_for_rw2

    def _get_tomita_path(self):
        tomita_path = (os.path.dirname(os.path.abspath(__file__)) +
                       "/../../tools/tomita/")

        if self.current_os == consts.OperatingSystem.LINUX_64:
            tomita_binary = "./tomita-linux64"
        elif self.current_os == consts.OperatingSystem.LINUX_32:
            tomita_binary = "./tomita-linux32"
        elif self.current_os == consts.OperatingSystem.WINDOWS:
            tomita_binary = "tomitaparser.exe"
        elif self.current_os == consts.OperatingSystem.MACOS:
            tomita_binary = "./tomita-mac"

        if not os.access(tomita_path + tomita_binary, os.F_OK):
            tomita_binary = None

        return tomita_path, tomita_binary

    def _calculate_word_frequencies(self, text):
        text = common_utils.prepare_text(text)
        words = common_utils.tokenize(text)
        res = collections.defaultdict(int)
        for word in words:
            res[word] += 1
        return res

    def _calculate_dt_frequencies(self, dependency_triples):
        res = collections.defaultdict(int)
        for dt in dependency_triples:
            res[dt] += 1
        return res

    def I(self, w1, r, w2):
        if (w1, r, w2) in self.I_memoized:
            return self.I_memoized[(w1, r, w2)]

        fr_w1rw2 = self.frequencies[w1, r, w2]
        if not fr_w1rw2:
            return 0.0
        fr__r_ = sum(self.frequencies[triple] for triple in self.dt_for_r[r])
        fr_w1r_ = sum(self.frequencies[triple] for triple in self.dt_for_w1r[(w1, r)])
        fr__rw2 = sum(self.frequencies[triple] for triple in self.dt_for_rw2[(r, w2)])
        res = max(math.log(float(fr_w1rw2) * fr__r_ / fr_w1r_ / fr__rw2), 0.0)
        self.I_memoized[(w1, r, w2)] = res
        return res

    def T(self, w):
        if w in self.T_memoized:
            return self.T_memoized[w]
        res = set(filter(lambda (r, w_): self.I(w, r, w_) > 0,
                         itertools.product(self.relations, self.words)))
        self.T_memoized[w] = res
        return res

    def similarity(self, w1, w2):
        numerator = sum(self.I(w1, r, w) + self.I(w2, r, w)
                        for (r, w) in self.T(w1) & self.T(w2))
        denominator = (sum(self.I(w1, r, w) for (r, w) in self.T(w1)) +
                       sum(self.I(w2, r, w) for (r, w) in self.T(w2)))
        if denominator:
            return numerator / denominator
        else:
            return 0.0

    def get_synonyms(self, threshold=0.3, return_similarity_measure=False):
        synonyms = collections.defaultdict(list)
        words = filter(lambda w: len(w) > 2 and
                                 self.word_frequencies[w] > self.number_of_texts / 50,
                       self.words)
        combs = itertools.combinations(words, 2)
        for w1, w2 in combs:
            sim = self.similarity(w1, w2)
            if sim > threshold:
                if return_similarity_measure:
                    synonyms[w1].append((w2, sim))
                    synonyms[w2].append((w1, sim))
                else:
                    synonyms[w1].append(w2)
                    synonyms[w2].append(w1)
        return synonyms



# -*- coding: utf-8 -*

import abc

from east.asts import base
from east.asts import utils
from east import consts

class AnnotatedSuffixTree(base.AST):
    __metaclass__ = abc.ABCMeta

    def __init__(self, strings_collection):
        super(AnnotatedSuffixTree, self).__init__(strings_collection)
        self.strings_collection = strings_collection
        self.root = self._construct(strings_collection)
        self._update_node_depth()

    def score(self, query, normalized=True, synonimizer=None):
        """
        Matches the string against the GAST using
        the algorithm described in [Chernyak, sections 1.3 & 1.4].
        
        Expects the input string to consist of
        alphabet letters only (no whitespaces etc.)
        
        Returns the score (a float in [0, 1]).
        
        query -- Unicode
        
        """
        
        query = query.replace(" ", "")
        result = 0
    
        # For each suffix of the string:
        for suffix_start in xrange(len(query)):
            
            suffix = query[suffix_start:]
            suffix_score = 0
            matched_chars = 0
            nodes_matched = 0
            
            child_node = self.root.chose_arc(suffix)
            while child_node:
                nodes_matched += 1
                (str_ind, substr_start, substr_end) = child_node.arc()
                match = utils.match_strings(
                            suffix, self.strings_collection[str_ind][substr_start:substr_end])
                suffix_score += child_node.conditional_probability()
                matched_chars += match
                suffix = suffix[match:]
                if suffix and match == substr_end-substr_start:
                    child_node = child_node.chose_arc(suffix)
                else:
                    break
            
            if matched_chars:
                if normalized:
                    suffix_result = (suffix_score + matched_chars - nodes_matched) / matched_chars
                else:
                    suffix_result = (suffix_score + matched_chars - nodes_matched)
                result += suffix_result
                    
            
        result /= len(query)
        return result

    def traverse_depth_first_pre_order(self, callback):
        """Traverses the annotated suffix tree in depth-first pre-order."""
        self.root.traverse_depth_first_pre_order(callback)

    def traverse_depth_first_post_order(self, callback):
        """Traverses the annotated suffix tree in depth-first post-order."""
        self.root.traverse_depth_first_post_order(callback)

    def traverse_breadth_first(self, callback):
        """Traverses the annotated suffix tree in breadth-first order."""
        self.root.traverse_breadth_first(callback, [])

    @abc.abstractmethod
    def _construct(self, strings_collection):
        """Constructs the annotated suffix tree and returns the pointer to its root."""
    
    def _update_node_depth(self):
        self.root.depth = 0
        def _calculate_depth(node):
            for k in node.children:
                node.children[k].depth = node.depth + 1
        self.traverse(_calculate_depth, consts.TraversalOrder.DEPTH_FIRST_PRE_ORDER)

    class Node:
        """
        Implementation of a Generalized Annotated Suffix Tree node.
        
        """ 
            
        def __init__(self):
            """ Hash table to store child nodes """
            self.children = {}
            """ Node weight """
            self.weight = 0
            """ Parent """
            self.parent = None
            """ Suffix link """
            self.suffix_link = None
            """ Arc that points to the node; Triple of form
                (string_index, substring_start_index[inclusive], substring_end_index[exclusive]) """
            self._arc = None
            """ Used in Ukkonen's algorithm """
            self._e = []
            """ Pointer to the strings collection """
            self.strings_collection = []
            # No sense in initializing node weight here
            # it is not possible to update it quickly while
            # constructing the tree; that's made later in one single pass
                
        
        def add_new_child(self, str_ind, substr_start, substr_end):
            """
            Creates and returns new child node.
            str_ind, substr_start, substr_end and parameters describe
            the substring that the new child should contain;
            for the given strings_collection that will be
            strings_collection[str_ind][substr_start:substr_end]
            
            """
            child_node = AnnotatedSuffixTree.Node()
            child_node.parent = self
            child_node.strings_collection = self.strings_collection
            child_node._arc = (str_ind, substr_start, substr_end)
            child_node._e = self._e
            self.children[self.strings_collection[str_ind][substr_start]] = child_node
            return child_node
        
        
        def add_child(self, child_node):
            """
            Adds an existing node as a new child
            for the current node.
            
            """
            (str_ind, substr_start, _) = child_node.arc()
            child_node.strings_collection = self.strings_collection
            self.children[self.strings_collection[str_ind][substr_start]] = child_node
            child_node.parent = self
        
        
        def remove_child(self, child_node):
            """
            Removes a child node from the current node.
            
            """
            (str_ind, substr_start, _) = child_node.arc()
            del self.children[self.strings_collection[str_ind][substr_start]]
            
            
        def conditional_probability(self):
            """
            Calculates the conditional probability
            of the first character in the node's substring;
            see [Chernyak, section 2.3] for details
            
            """
            return float(self.weight) / self.parent.weight
            
            
        def arc(self):
            """
            Returns a tuple of form (str_ind, substr_start, substr_end)
            that describes the substring that the current node contains
            (can be also imagined as the label of the arc that poins to this node).
            For the given strings_collection that will be
            strings_collection[str_ind][substr_start:substr_end]
            
            """
            (str_ind, substr_start, substr_end) = self._arc
            if substr_end == -1:
                substr_end = self._e[str_ind]
            return (str_ind, substr_start, substr_end)
        
        
        def arc_label(self):
            (str_ind, substr_start, substr_end) = self.arc()
            return self.strings_collection[str_ind][substr_start:substr_end]
        
            
        def chose_arc(self, string):
            """
            Returns the child node, the arc to which is
            labeled with a string that starts with the same
            character as the 'string' parameter.
            Returns None if no such arc exisits.
            O(1) amortized time complexity
            (since we use hash tables for storing children).
            
            """
            if string[0] in self.children:
                return self.children[string[0]]
            elif not string[0]:
                return self
                
            return None
        
        
        def is_leaf(self):
            """
            Returns whether the node is a leaf.
            
            """
            return not self.children and not self.is_root()
            
            
        def is_root(self):
            """
            Returns whether the node is a root of the tree.
            
            """
            return not self.parent
        
        
        def path(self):
            """
            Returns a string that represents the path to the current node.
            
            """
            res = ''
            node = self
            while not node.is_root():
                (str_ind, substr_start, substr_end) = node.arc()
                res = self.strings_collection[str_ind][substr_start:substr_end] + res
                node = node.parent
            return res
        
        
        def equals(self, other):
            """
            Determines whether the current node equals to the other node;
            that is, whether they have the same weight and equal children.
            
            """
            if self.weight != other.weight:
                return False
            
            if set(self.children.keys()) != set(other.children.keys()):
                return False
                
            for k in self.children.keys():
                if not self.children[k].equals(other.children[k]):
                    return False
                
            return True
            
            
        #######################################################
        ######             T R A V E R S A L S           ######
        #######################################################
        
        def traverse_depth_first_pre_order(self, callback):
            """
            Traverses the tree in depth-first top-down order,
            calling the callback function in each node.
            The callback function should take the node as its parameter.
    
            """
            callback(self)
            for k in self.children:
                self.children[k].traverse_depth_first_pre_order(callback)
                
        def traverse_depth_first_post_order(self, callback):
            """
            Traverses the tree in depth-first bottom-up order,
            calling the callback function in each node.
            The callback function should take the node as its parameter.
    
            """
            for k in self.children:
                self.children[k].traverse_depth_first_post_order(callback)
            callback(self)
        
        def traverse_breadth_first(self, callback, queue):
            """
            Traverses the tree in breadth-first top-down order,
            calling the callback function in each node.
            The callback function should take the node as its parameter.
    
            """
            callback(self)
            for k in self.children:
                queue.append(self.children[k])
            if queue:
                queue[0].traverse_breadh_first_top_down(callback, queue[1:])
            
            
        def __str__(self): 
            """
            Returns just the node annotation (for networkx graph drawing)
            
            """
            return str(self.weight)
            


# -*- coding: utf-8 -*

import gc
import getopt
import sys
import time

from analysis import utils
from east.asts import base

def main(args):
    opts, args = getopt.getopt(args, "")

    n_from = int(args[0])
    n_to = int(args[1])
    n_step = int(args[2]) if len(args) >= 3 else 1
    m = int(args[3]) if len(args) >= 4 else 100

    repeats = 5  # for each n

    for ast_algorithm in ["ast_naive", "ast_linear", "easa"]:
        print ast_algorithm
        for n in xrange(n_from, n_to + 1, n_step):
            t = 0
            for _ in xrange(repeats):
                strings_collection = utils.worst_case_strings_collection(m, n)
                start = time.clock()
                base.AST.get_ast(strings_collection, ast_algorithm)
                t += time.clock() - start
            gc.collect()
            print("%i\t%.2f" % (n, t / repeats))
        print ""


if __name__ == '__main__':
    main(sys.argv[1:])



# -*- coding: utf-8 -*

import itertools
import testtools

from east.asts import base


class BASEAstTestCase(testtools.TestCase):

    def setUp(self):
        super(BASEAstTestCase, self).setUp()
        self.strings_collection = ["abcd efg ops", "xyzq", "test"]
        self.queries = ["aqcb", "efgp", "mn4"]

    def test_matching_scores_equality(self):
        algorithms = ["easa", "ast_linear", "ast_naive"]
        for normalized in [True, False]:
            for alg1, alg2 in itertools.combinations(algorithms, 2):
                ast1 = base.AST.get_ast(self.strings_collection, alg1)
                ast2 = base.AST.get_ast(self.strings_collection, alg2)
                for query in self.queries:
                    self.assertEqual(ast1.score(query, normalized=normalized),
                                     ast2.score(query, normalized=normalized))



# -*- coding: utf-8 -*

import testtools

from east.asts import utils


class AstUtilsTestCase(testtools.TestCase):

    def test_match_strings_empty(self):
        self.assertEqual(utils.match_strings("abc", "bc"), 0)
        self.assertEqual(utils.match_strings("", ""), 0)

    def test_match_strings_partial(self):
        self.assertEqual(utils.match_strings("abc", "ac"), 1)
        self.assertEqual(utils.match_strings("mnc", "mnd"), 2)

    def test_match_strings_full(self):
        self.assertEqual(utils.match_strings("abc", "abc"), 3)
        self.assertEqual(utils.match_strings("abc", "abcd"), 3)

    def test_index_int(self):
        self.assertEqual(utils.index([0, 2, 4, 6], 0), 0)
        self.assertEqual(utils.index([0, 2, 4, 6], 4), 2)
        self.assertEqual(utils.index([0, 2, 4, 6], 6), 3)

    def test_index_string(self):
        self.assertEqual(utils.index(["a", "b", "c", "d"], "a"), 0)
        self.assertEqual(utils.index(["a", "b", "c", "d"], "c"), 2)
        self.assertEqual(utils.index(["a", "b", "c", "d"], "d"), 3)



# -*- coding: utf-8 -*

import itertools
import sys

from east.asts import base
from east import utils

def keyphrases_table(keyphrases, texts, ast_algorithm="easa", normalized=True, synonimizer=None):
    """
    Constructs the keyphrases table, containing their matching scores in a set of texts.

    The resulting table is stored as a dictionary of dictionaries,
    where the entry table["keyphrase"]["text"] corresponds
    to the matching score (0 <= score <= 1) of keyphrase "keyphrase"
    in the text named "text".
    
    :param keyphrases: list of unicode strings
    :param texts: dictionary of form {text_name: text}
    :param ast_algorithm: AST implementation to use
    :param normalized: whether the scores should be normalized
    :param synonimizer: SynonymExtractor object to be used

    :returns: dictionary of dictionaries, having keyphrases on its first level and texts
              on the second level.  
    """

    i = 0
    total_texts = len(texts)
    asts = {}
    for text in texts:
        i += 1
        if not utils.output_is_redirected():
            sys.stdout.write("\rConstructing ASTs: %i/%i" % (i, total_texts))
            sys.stdout.flush()
        asts[text] = base.AST.get_ast(utils.text_to_strings_collection(texts[text]), ast_algorithm)

    i = 0
    total_keyphrases = len(keyphrases)
    total_scores = total_texts * total_keyphrases
    res = {}
    for keyphrase in keyphrases:
        res[keyphrase] = {}
        for text in texts:
            i += 1
            if not utils.output_is_redirected():
                sys.stdout.write("\rCalculating matching scores: %i/%i" % (i, total_scores))
                sys.stdout.flush()
            res[keyphrase][text] = asts[text].score(keyphrase, normalized=normalized,
                                                    synonimizer=synonimizer)

    if not utils.output_is_redirected():
        sys.stdout.write("\r" + " " * 80 + "\r")
        sys.stdout.flush()

    return res


def keyphrases_graph(keyphrases, texts, significance_level=0.6, score_treshold=0.25,
                     ast_algorithm="easa", normalized=True, synonimizer=None):
    """
    Constructs the keyphrases relation graph based on the given texts corpus.

    The graph construction algorithm is based on the analysis of co-occurrences of key phrases
    in the text corpus. A key phrase is considered to imply another one if that second phrase
    occurs frequently enough in the same texts as the first one (that frequency is controlled
    by the significance_level). A keyphrase counts as occuring in a text if its matching score
    for that text ecxeeds some threshold (Mirkin, Chernyak, & Chugunova, 2012).

    :param keyphrases: list of unicode strings
    :param texts: dictionary of form {text_name: text}
    :param significance_level: significance level of the graph in [0; 1], 0.6 by default
    :param score_treshold: threshold for the matching score in [0; 1] where a keyphrase starts
                           to be considered as occuring in the corresponding text, 0.25 by default
    :param synonimizer: SynonymExtractor object to be used

    :returns: graph in a dictionary format: dictionary keys are node labels, while each key points
              to a list of adjacent node labels ({"A": ["B", "C"], "B": ["A"], "C": []})
    """

    # Keyphrases table
    table = keyphrases_table(keyphrases, texts, ast_algorithm, normalized, synonimizer)
    
    # Dictionary { "keyphrase" => set(names of texts containing "keyphrase") }
    keyphrase_texts = {keyphrase: set([text for text in texts
                                       if table[keyphrase][text] >= score_treshold])
                       for keyphrase in keyphrases}
    
    # Initializing the graph object with nodes
    graph = {k: [] for k in keyphrases}
    
    # Creating arcs
    # NOTE(msdubov): permutations(), unlike combinations(), treats (1,2) and (2,1) as different
    for (k1, k2) in itertools.permutations(keyphrases, 2):
        if (len(keyphrase_texts[k1]) > 0 and 
            float(len(keyphrase_texts[k1] & keyphrase_texts[k2])) /
            len(keyphrase_texts[k1]) >= significance_level):
            graph[k1].append(k2)
            
    return graph



import setuptools

setuptools.setup(
    name = "EAST",
    packages = setuptools.find_packages(),
    version = "0.2.1",
    description = "Text analysis library based on the Annotated Suffix Tree method",

    install_requires = [
        "numpy>=1.7.1",
        "docutils>=0.3",
        "testtools>=0.9.35"
    ],

    entry_points = {
        "console_scripts": [
            "east = east.main:main"
        ]
    },

    author = "Mikhail Dubov",
    author_email = "msdubov@gmail.com",
    license = "MIT",
    url = "https://github.com/msdubov/AST-text-analysis",
    download_url = "https://github.com/msdubov/AST-text-analysis/tarball/0.2.1",
    keywords = ["text analysis", "suffix tree", "synonym extraction"]
)



# -*- coding: utf-8 -*

def table2xml(keyphrases_table):
    res = u"<table>\n"
    for keyphrase in sorted(keyphrases_table.keys()):
        res += u'  <keyphrase value="%s">\n' % keyphrase
        for text in sorted(keyphrases_table[keyphrase].keys()):
            res += u'    <text name="%s">' % text
            res += u'%.3f' % keyphrases_table[keyphrase][text]
            res += u'</text>\n'
        res += u'  </keyphrase>\n'
    res += u"</table>\n"
    return res


def table2csv(keyphrases_table):

    def quote(s):
        return '"' + s.replace('"', "'") + '"'

    keyphrases = sorted(keyphrases_table.keys())
    texts = sorted(keyphrases_table[keyphrases[0]].keys())
    res = u"," + ",".join(map(quote, keyphrases)) + "\n"  # Heading
    for text in texts:
        scores = map(lambda score: u"%.3f" % score,
                     [keyphrases_table[keyphrase][text] for keyphrase in keyphrases])
        res += (quote(text) + "," + ",".join(scores) + "\n")
    return res


def graph2edges(graph):
    res = u""
    for node in graph:
        if graph[node]:
            res += "%s -> %s\n" % (node, ", ".join(graph[node]))
    return res


def graph2gml(graph):
    res = u"graph\n[\n"
    i = 0
    node_ids = {}
    for node in graph:
        res += '  node\n  [\n    id %i\n    label "%s"\n  ]\n' % (i, node)
        node_ids[node] = i
        i += 1
    for node in graph:
        for adj_node in graph[node]:
            res += '  edge\n  [\n    source %i\n    target %i\n  ]\n' % (node_ids[node],
                                                                         node_ids[adj_node])
    res += "]\n"
    return res
